#include "rei-json/Array.h"
#include "rei-json/Field.h"
#include "rei-json/json.h"
#include <cassert>
#include <cctype>
#include <charconv>
#include <print>
#include <string>
#include <string_view>
#include <utility>

namespace rei::json {

	class StringParser {

		public:

		StringParser(const std::string& str) : str(str) {  }

		int currentLine() const {
			return line;
		}

		int currentColumn() const {
			return i - lineStartingIndex;
		}

		const char& current() const {
			return str[i];
		}

		bool is(const char& c) const {
			return str[i] == c;
		}

		bool match(const char* c, int length) const {
			return std::string_view{str}.substr(i, length) == c;
		}

		bool isDigit() const {
			return isdigit(str[i]);
		}

		void skipWhitespaces() {
			while (str[i] == ' ' || str[i] == '\t' || str[i] == '\n' || str[i] == '\v') {
				if (str[i] == '\n') {
					++line;
					lineStartingIndex = i + 1;
				}
				++i;
			}
		}

		void goToNextDoubleQuote() {
			while (str[i] != '"') {
				if (str[i] == '\\') {
					++(i);
				}
				++(i);
			}
		}

		void skipDigits() {
			while (isdigit(str[i]) || str[i] == '-') {
				++(i);
			}
		}

		void nextChar(int count = 1) {
			i += count;
		}

		int currentIndex() const {
			return i;
		}

		std::string substr(int start, int length) const {
			return str.substr(start, length);
		}

		const char *c_str() const {
			return str.c_str();
		}

		FieldValue parseField(StringParser& parser) {
			if (parser.is('"')) {
				parser.nextChar();
				const int valueStart = parser.currentIndex();
				parser.goToNextDoubleQuote();
				const int valueLength = parser.currentIndex() - valueStart;
				parser.nextChar();
				return FieldValue{parser.substr(valueStart, valueLength)};
			} else if (parser.isDigit() || parser.is('-')) {
				const int valueStart = parser.currentIndex();
				parser.skipDigits();
				const int valueLength = parser.currentIndex() - valueStart;
				int value;
				std::from_chars(parser.c_str() + valueStart, parser.c_str() + valueStart + valueLength, value);
				return FieldValue{value};
			} else if (parser.match("true", 4)) {
				parser.nextChar(4);
				return FieldValue{true};
			} else if (parser.match("false", 5)) {
				parser.nextChar(5);
				return FieldValue{false};
			} else if (parser.match("null", 4)) {
				parser.nextChar(4);
				return FieldValue{FieldType::Null};
			} else if (parser.is('{')){
				JsonObject newObject{}; 
				parseObject(parser, &newObject);
				return FieldValue{std::move(newObject)};
			} else if (parser.is('[')){
				JsonArray newArray{}; 
				parseArray(parser, &newArray);
				return FieldValue{std::move(newArray)};
			}
			throw ParsingError("Invalid JSON value", currentLine(), currentColumn());
		}

		void parseObjectField(StringParser& parser, JsonObject* object) {
			assert(parser.is('"'));
			parser.nextChar();
			const int keyStart = parser.currentIndex();
			parser.goToNextDoubleQuote();
			const int keyLength = parser.currentIndex() - keyStart;
			const std::string key = parser.substr(keyStart, keyLength);
			parser.nextChar();
			parser.skipWhitespaces();
			if (!parser.is(':')) {
				throw ParsingError("Expected ':'", currentLine(), currentColumn());
			}
			parser.nextChar();
			parser.skipWhitespaces();
			object->set(key, parseField(parser));
		}

		void parseArrayField(StringParser& parser, JsonArray* array) {
			parser.skipWhitespaces();
			array->push(parseField(parser));
		}

		void parseArray(StringParser& parser, JsonArray* object) {
			assert(parser.is('['));
			parser.nextChar();
			parser.skipWhitespaces();
			while (!parser.is(']')) {
				parseArrayField(parser, object);
				parser.skipWhitespaces();
				if (parser.is(',')) {
					parser.nextChar();
					parser.skipWhitespaces();
					if (parser.is(']')) {
						throw ParsingError("Found extra comma (',')", currentLine(), currentColumn());
					}
				} else if (!parser.is(']')) {
					throw ParsingError("Expected comma (',') or array ending (']')", currentLine(), currentColumn());
				}
			}
			parser.nextChar();
		}
		
		void parseObject(StringParser& parser, JsonObject* object) {
			assert(parser.is('{'));
			parser.nextChar();
			parser.skipWhitespaces();
			while (!parser.is('}')) {
				if (!parser.is('"')) {
					throw ParsingError("Expected '\"'", currentLine(), currentColumn());
				}
				parseObjectField(parser, object);
				parser.skipWhitespaces();
				if (parser.is(',')) {
					parser.nextChar();
					parser.skipWhitespaces();
					if (parser.is('}')) {
						throw ParsingError("Found extra comma (',')", currentLine(), currentColumn());
					}
				} else if (!parser.is('}')) {
					throw ParsingError("Expected comma (',') or object ending ('}')", currentLine(), currentColumn());
				}
			}
			parser.nextChar();
		}

		private:

		unsigned i = 0;
		unsigned line = 1;
		unsigned lineStartingIndex = 1;
		const std::string str;
	};

	FieldValue parse(const std::string& jsonStr) {
		StringParser parser{jsonStr};

		parser.skipWhitespaces();
		if (parser.is('{')) {
			JsonObject object{};
			parser.parseObject(parser, &object);
			return std::move(object);
		} else if (parser.is('[')) {
			JsonArray array{};
			parser.parseArray(parser, &array);
			return std::move(array);
		}
		throw ParsingError("Expected JSON object or array", parser.currentLine(), parser.currentColumn());
	}
}
